import json

from collections import defaultdict
from datetime import datetime

from django.db import connection
from django.http import HttpResponse

from django.shortcuts import get_object_or_404

from catmaid.models import *
from catmaid.fields import Double3D
from catmaid.control.authentication import *
from catmaid.control.common import *

import sys
try:
    import networkx as nx
except:
    pass


@requires_user_role([UserRole.Annotate, UserRole.Browse])
def node_list_tuples(request, project_id=None):
    ''' Retrieve an JSON array with four entries:
    [0] an array of arrays, each array representing a treenode
    [1] an array of arrays, each array representing a connector and containing
    arrays inside that specify the relations between the connector and treenodes.
    In this function tuples are used as much as possible for immutable list,
    and uses directly the tuples returned by the database cursor.
    [2] the labels, if requested.
    [3] a boolean which is true when the node limit has been reached.
    The returned JSON data is therefore sensitive to indices in the array,
    so care must be taken never to alter the order of the variables in the SQL
    statements without modifying the accesses to said data both in this function
    and in the client that consumes it.
    '''
    params = {}
    # z: the section index in calibrated units.
    # width: the width of the field of view in calibrated units.
    # height: the height of the field of view in calibrated units.
    # zres: the resolution in the Z axis, used to determine the thickness of a section.
    # as: the ID of the active skeleton
    # top: the Y coordinate of the bounding box (field of view) in calibrated units
    # left: the X coordinate of the bounding box (field of view) in calibrated units
    atnid = int(request.POST.get('atnid', -1))
    for p in ('top', 'left', 'z', 'width', 'height', 'zres'):
        params[p] = float(request.POST.get(p, 0))
    params['limit'] = 5000  # Limit the number of retrieved treenodes within the section
    params['project_id'] = project_id
    
    relation_map = get_relation_to_id_map(project_id)
    class_map = get_class_to_id_map(project_id)

    if 'skeleton' not in class_map:
        raise Exception('Can not find "skeleton" class for this project')

    for relation in ['presynaptic_to', 'postsynaptic_to', 'model_of']:
        if relation not in relation_map:
            raise Exception('Can not find "%s" relation for this project' % relation)

    try:
        cursor = connection.cursor()
        # Fetch treenodes which are in the bounding box,
        # which in z it includes the full thickess of the prior section
        # and of the next section (therefore the '<' and not '<=' for zhigh)
        response_on_error = 'Failed to query treenodes'
        params['bottom'] = params['top'] + params['height']
        params['right'] = params['left'] + params['width']
        cursor.execute('''
        SELECT
            t1.id,
            t1.parent_id,
            (t1.location).x AS x,
            (t1.location).y AS y,
            (t1.location).z AS z,
            t1.confidence,
            t1.radius,
            t1.skeleton_id,
            t1.user_id,
            t2.id,
            t2.parent_id,
            (t2.location).x AS x,
            (t2.location).y AS y,
            (t2.location).z AS z,
            t2.confidence,
            t2.radius,
            t2.skeleton_id,
            t2.user_id
        FROM treenode t1
             INNER JOIN treenode t2 ON
               (   (t1.id = t2.parent_id OR t1.parent_id = t2.id)
                OR (t1.parent_id IS NULL AND t1.id = t2.id))
        WHERE
            (t1.location).z = %(z)s
            AND (t1.location).x > %(left)s
            AND (t1.location).x < %(right)s
            AND (t1.location).y > %(top)s
            AND (t1.location).y < %(bottom)s
            AND t1.project_id = %(project_id)s
        LIMIT %(limit)s
        ''', params)

        # Above, notice that the join is done for:
        # 1. A parent-child or child-parent pair (where the first one is in section z)
        # 2. A node with itself when the parent is null
        # This is by far the fastest way to retrieve all parents and children nodes
        # of the nodes in section z within the specified 2d bounds.

        # A list of tuples, each tuple containing the selected columns for each treenode
        # The id is the first element of each tuple
        treenodes = []
        # A set of unique treenode IDs
        treenode_ids = set()

        is_superuser = request.user.is_superuser
        user_id = request.user.id

        n_retrieved_nodes = 0 # at one per row, only those within the section
        for row in cursor.fetchall():
          n_retrieved_nodes += 1
          t1id = row[0]
          if t1id not in treenode_ids:
              treenode_ids.add(t1id)
              treenodes.append(row[0:8] + (is_superuser or row[8] == user_id,))
          t2id = row[9]
          if t2id not in treenode_ids:
              treenode_ids.add(t2id)
              treenodes.append(row[9:17] + (is_superuser or row[17] == user_id,))

        # Find connectors related to treenodes in the field of view
        # Connectors found attached to treenodes
        crows = []

        if treenode_ids:
            response_on_error = 'Failed to query connector locations.'
            cursor.execute('''
            SELECT connector.id AS id,
                (connector.location).x AS x,
                (connector.location).y AS y,
                (connector.location).z AS z,
                connector.confidence AS confidence,
                treenode_connector.relation_id AS treenode_relation_id,
                treenode_connector.treenode_id AS tnid,
                treenode_connector.confidence AS tc_confidence,
                connector.user_id AS user_id
            FROM treenode
                 INNER JOIN treenode_connector ON treenode.id = treenode_connector.treenode_id
                 INNER JOIN connector ON treenode_connector.connector_id = connector.id
            WHERE treenode.id IN (%s)''' % ','.join(str(x) for x in treenode_ids))

            for row in cursor.fetchall():
                crows.append(row)
        
        # Obtain connectors within the field of view that were not captured above.
        # Uses a LEFT OUTER JOIN to include disconnected connectors,
        # that is, connectors that aren't referenced from treenode_connector.
        cursor.execute('''
        SELECT connector.id AS id,
            (connector.location).x AS x,
            (connector.location).y AS y,
            (connector.location).z AS z,
            connector.confidence AS confidence,
            treenode_connector.relation_id AS treenode_relation_id,
            treenode_connector.treenode_id AS tnid,
            treenode_connector.confidence AS tc_confidence,
            connector.user_id AS user_id
        FROM connector LEFT OUTER JOIN treenode_connector
                       ON connector.id = treenode_connector.connector_id
        WHERE connector.project_id = %(project_id)s
          AND (connector.location).z = %(z)s
          AND (connector.location).x > %(left)s
          AND (connector.location).x < %(right)s
          AND (connector.location).y > %(top)s
          AND (connector.location).y < %(bottom)s
        ''', params)

        for row in cursor.fetchall():
            crows.append(row)

        connectors = []
        # A set of missing treenode IDs
        missing_treenode_ids = set()
        # Check if the active treenode is present; if not, load it
        if -1 != atnid and atnid not in treenode_ids:
            # If atnid is a connector, it doesn't matter, won't be found in treenode table
            missing_treenode_ids.add(atnid)
        # A set of unique connector IDs
        connector_ids = set()
        # The relations between connectors and treenodes, stored
        # as connector ID keys vs a list of tuples, each with the treenode id,
        # the type of relation (presynaptic_to or postsynaptic_to), and the confidence.
        # The list of tuples is generated later from a dict,
        # so that repeated tnid entries are overwritten.
        pre = defaultdict(dict)
        post = defaultdict(dict)

        # Process crows (rows with connectors) which could have repeated connectors
        # given the join with treenode_connector
        presynaptic_to = relation_map['presynaptic_to']
        for row in crows:
            # Collect treeenode IDs related to connectors but not yet in treenode_ids
            # because they lay beyond adjacent sections
            tnid = row[6] # The tnid column is index 7 (see SQL statement above)
            cid = row[0] # connector ID
            if tnid is not None:
                if tnid not in treenode_ids:
                    missing_treenode_ids.add(tnid)
                # Collect relations between connectors and treenodes
                # row[5]: treenode_relation_id
                # row[6]: treenode_id (tnid above)
                # row[7]: tc_confidence
                if row[5] == presynaptic_to:
                    pre[cid][tnid] = row[7]
                else:
                    post[cid][tnid] = row[7]

            # Collect unique connectors
            if cid not in connector_ids:
                connectors.append(row)
                connector_ids.add(cid)

        # Fix connectors to contain only the relevant entries, plus the relations
        for i in xrange(len(connectors)):
            c = connectors[i]
            cid = c[0]
            connectors[i] = (cid, c[1], c[2], c[3], c[4],
                    [kv for kv in  pre[cid].iteritems()],
                    [kv for kv in post[cid].iteritems()],
                    is_superuser or c[8] == user_id)


        # Fetch missing treenodes. These are related to connectors
        # but not in the bounding box of the field of view.
        # This is so that we can draw arrows from any displayed connector
        # to all of its connected treenodes, even if one is several slices
        # below.

        if missing_treenode_ids:
            params['missing'] = tuple(missing_treenode_ids)
            response_on_error = 'Failed to query treenodes from connectors'
            cursor.execute('''
            SELECT id,
                parent_id,
                (location).x AS x,
                (location).y AS y,
                (location).z AS z,
                confidence,
                radius,
                skeleton_id,
                user_id
            FROM treenode
            WHERE id IN %(missing)s''', params)

            for row in cursor.fetchall():
                treenodes.append(row)
                treenode_ids.add(row[0:8] + (is_superuser or row[8] == user_id,))

        labels = defaultdict(list)
        if request.POST['labels']:
            z0 = params['z']
            # Collect treenodes visible in the current section
            visible = ','.join(str(row[0]) for row in treenodes if row[4] == z0)
            if visible:
                cursor.execute('''
                SELECT treenode.id, class_instance.name
                FROM treenode, class_instance, treenode_class_instance, relation
                WHERE relation.id = treenode_class_instance.relation_id
                  AND relation.relation_name = 'labeled_as'
                  AND treenode_class_instance.treenode_id = treenode.id
                  AND class_instance.id = treenode_class_instance.class_instance_id
                  AND treenode.id IN (%s)
                ''' % visible)
                for row in cursor.fetchall():
                    labels[row[0]].append(row[1])

            # Collect connectors visible in the current section
            visible = ','.join(str(row[0]) for row in connectors if row[3] == z0)
            if visible:
                cursor.execute('''
                SELECT connector.id, class_instance.name
                FROM connector, class_instance, connector_class_instance, relation
                WHERE relation.id = connector_class_instance.relation_id
                  AND relation.relation_name = 'labeled_as'
                  AND connector_class_instance.connector_id = connector.id
                  AND class_instance.id = connector_class_instance.class_instance_id
                  AND connector.id IN (%s)
                ''' % visible)
                for row in cursor.fetchall():
                    labels[row[0]].append(row[1])

        return HttpResponse(json.dumps((treenodes, connectors, labels, n_retrieved_nodes == params['limit'])))

    except Exception as e:
        raise Exception(response_on_error + ':' + str(e))


@requires_user_role(UserRole.Annotate)
def update_location_reviewer(request, project_id=None, node_id=None):
    """ Updates the reviewer id and review time of a node """
    p = get_object_or_404(Project, pk=project_id)
    loc = Location.objects.get(
        pk=node_id,
        project=p)
    loc.reviewer_id=request.user.id
    loc.review_time=datetime.now()
    loc.save()
    return HttpResponse(json.dumps({'reviewer_id': request.user.id}), mimetype='text/json')


@requires_user_role(UserRole.Annotate)
def update_confidence(request, project_id=None, node_id=0):
    tnid = int(node_id)
    can_edit_or_fail(request.user, tnid, 'treenode')

    new_confidence = int(request.POST.get('new_confidence', 0))
    if new_confidence < 1 or new_confidence > 5:
        return HttpResponse(json.dumps({'error': 'Confidence not in range 1-5 inclusive.'}))
    to_connector = request.POST.get('to_connector', 'false') == 'true'
    if to_connector:
        # Could be more than one. The GUI doesn't allow for specifying to which one.
        rows_affected = TreenodeConnector.objects.filter(treenode=tnid).update(confidence=new_confidence)
    else:
        rows_affected = Treenode.objects.filter(id=tnid).update(confidence=new_confidence,editor=request.user)

    if rows_affected > 0:
        location = Location.objects.get(id=tnid).location
        insert_into_log(project_id, request.user.id, "change_confidence", location, "Changed to %s" % new_confidence)
        return HttpResponse(json.dumps({'message': 'success'}), mimetype='text/json')

    # Else, signal error
    if to_connector:
        return HttpResponse(json.dumps({'error': 'Failed to update confidence between treenode %s and connector.' % tnid}))
    else:
        return HttpResponse(json.dumps({'error': 'Failed to update confidence at treenode %s.' % tnid}))



@requires_user_role([UserRole.Annotate, UserRole.Browse])
def most_recent_treenode(request, project_id=None):
    treenode_id = int(request.POST.get('treenode_id', -1))
    skeleton_id = Treenode.objects.get(pk=treenode_id).skeleton_id

    try:
        # Select the most recently edited node
        tn = Treenode.objects\
             .filter(project=project_id,
            skeleton=skeleton_id,
            editor=request.user)\
             .extra(select={'most_recent': 'greatest(treenode.creation_time, treenode.edition_time)'})\
             .extra(order_by=['-most_recent', '-treenode.id'])[0] # [0] generates a LIMIT 1
    except IndexError:
        return HttpResponse(json.dumps({'error': 'No skeleton and neuron found for treenode %s' % treenode_id}))

    return HttpResponse(json.dumps({
        'id': tn.id,
        #'skeleton_id': tn.skeleton.id,
        'x': int(tn.location.x),
        'y': int(tn.location.y),
        'z': int(tn.location.z),
        #'most_recent': str(tn.most_recent) + tn.most_recent.strftime('%z'),
        #'most_recent': tn.most_recent.strftime('%Y-%m-%d %H:%M:%S.%f'),
        #'type': 'treenode'
    }))


@requires_user_role(UserRole.Annotate)
def node_update(request, project_id=None):
    nodes = {}
    for key, value in request.POST.items():
        # TODO wtf? The encoding in overlay.js/updateNodePositions is wasteful and arcane for no reason. JSON can encode javascript objects just fine.
        parsed_key = re.search('^(?P<property>[a-zA-Z_]+)(?P<node_index>[0-9]+)$', key)
        if not parsed_key:
            continue
        node_index = parsed_key.group('node_index')
        node_property = parsed_key.group('property')
        if node_index not in nodes:
            nodes[node_index] = {}
        nodes[node_index][node_property] = value

    required_properties = set( ['node_id', 'x', 'y', 'z', 'type'] )
    now = datetime.now()
    for node_index, node in nodes.items():
        for req_prop in required_properties:
            if req_prop not in node:
                raise Exception('Missing key: %s in index %s' % (req_prop, node_index))

        try:
            if node['type'] == 'treenode':
                try:
                    can_edit_or_fail(request.user, node['node_id'], 'treenode')
                except ObjectDoesNotExist:
                    # Ignore deleted objects.
                    # This can happen because of the async between the client and the server: the client may move the node in the same mouse action that triggers the deletion of the node.
                    continue
                Treenode.objects.filter(id=node['node_id']).update(
                    editor=request.user,
                    edition_time=now,
                    location=Double3D(float(node['x']), float(node['y']), float(node['z'])))
            elif node['type'] == 'connector':
                try:
                    can_edit_or_fail(request.user, node['node_id'], 'connector')
                except ObjectDoesNotExist:
                    # Ignore deleted objects.
                    continue
                Connector.objects.filter(id=node['node_id']).update(
                    editor=request.user,
                    edition_time=now,
                    location=Double3D(float(node['x']), float(node['y']), float(node['z'])))
            else:
                raise Exception('Unknown node type: %s' % node['type'])
        except:
            raise Exception('Failed to update treenode: %s' % node['node_id'])

    return HttpResponse(json.dumps({'updated': len(nodes)}))


@requires_user_role([UserRole.Annotate, UserRole.Browse])
def node_nearest(request, project_id=None):
    params = {}
    param_float_defaults = {
        'x': 0,
        'y': 0,
        'z': 0}
    param_int_defaults = {
        'skeleton_id': -1,
        'neuron_id': -1}
    for p in param_float_defaults.keys():
        params[p] = float(request.POST.get(p, param_float_defaults[p]))
    for p in param_int_defaults.keys():
        params[p] = int(request.POST.get(p, param_int_defaults[p]))
    relation_map = get_relation_to_id_map(project_id)

    if params['skeleton_id'] < 0 and params['neuron_id'] < 0:
        raise Exception('You must specify either a skeleton or a neuron')

    for rel in ['part_of', 'model_of']:
        if rel not in relation_map:
            raise Exception('Could not find required relation %s for project %s.' % (rel, project_id))

    skeletons = []
    if params['skeleton_id'] > 0:
        skeletons.append(params['skeleton_id'])

    response_on_error = ''
    try:
        if params['neuron_id'] > 0:  # Add skeletons related to specified neuron
            # Assumes that a cici 'model_of' relationship always involves a
            # skeleton as ci_a and a neuron as ci_b.
            response_on_error = 'Finding the skeletons failed.'
            neuron_skeletons = ClassInstanceClassInstance.objects.filter(
                class_instance_b=params['neuron_id'],
                relation=relation_map['model_of'])
            for neur_skel_relation in neuron_skeletons:
                skeletons.append(neur_skel_relation.class_instance_a_id)

        # Get all treenodes connected to skeletons
        response_on_error = 'Finding the treenodes failed.'
        treenodes = Treenode.objects.filter(project=project_id, skeleton__in=skeletons)

        def getNearestTreenode(x, y, z, treenodes):
            minDistance = -1
            nearestTreenode = None
            for tn in treenodes:
                xdiff = x - tn.location.x
                ydiff = y - tn.location.y
                zdiff = z - tn.location.z
                distanceSquared = xdiff ** 2 + ydiff ** 2 + zdiff ** 2
                if distanceSquared < minDistance or minDistance < 0:
                    nearestTreenode = tn
                    minDistance = distanceSquared
            return nearestTreenode

        nearestTreenode = getNearestTreenode(
            params['x'],
            params['y'],
            params['z'],
            treenodes)
        if nearestTreenode is None:
            raise Exception('No treenodes were found for skeletons in %s' % skeletons)

        return HttpResponse(json.dumps({
            'treenode_id': nearestTreenode.id,
            'x': int(nearestTreenode.location.x),
            'y': int(nearestTreenode.location.y),
            'z': int(nearestTreenode.location.z),
            'skeleton_id': nearestTreenode.skeleton_id}))

    except Exception as e:
        raise Exception(response_on_error + ':' + str(e))


def _skeleton_as_graph(skeleton_id):
    # Fetch all nodes of the skeleton
    cursor = connection.cursor()
    cursor.execute('''
        SELECT id, parent_id
        FROM treenode
        WHERE skeleton_id=%s''', [skeleton_id])
    # Create a directed graph of the skeleton
    graph = nx.DiGraph()
    for row in cursor.fetchall():
        # row[0]: id
        # row[1]: parent_id
        graph.add_node(row[0])
        if row[1]:
            # Create directional edge from parent to child
            graph.add_edge(row[1], row[0])
    return graph


def _fetch_location(treenode_id):
    cursor = connection.cursor()
    cursor.execute('''
        SELECT
          id,
          (location).x AS x,
          (location).y AS y,
          (location).z AS z,
          skeleton_id
        FROM treenode
        WHERE id=%s''', [treenode_id])
    return cursor.fetchone()


def _fetch_location_connector(connector_id):
    cursor = connection.cursor()
    cursor.execute('''
        SELECT
          id,
          (location).x AS x,
          (location).y AS y,
          (location).z AS z
        FROM connector
        WHERE id=%s''', [connector_id])
    return cursor.fetchone()


@requires_user_role([UserRole.Annotate, UserRole.Browse])
def get_location(request, project_id=None):
    try:
        tnid = int(request.POST['tnid'])
        nodetype = request.POST.get('type', 'treenode')
        if nodetype == 'connector':
            return HttpResponse(json.dumps(_fetch_location_connector(tnid)))
        else:
            return HttpResponse(json.dumps(_fetch_location(tnid)))
    except Exception as e:
        raise Exception('Could not obtain the location of node with id #%s' % tnid)


def _find_first_interesting_node(sequence):
    """ Find the first node that:
    1. Has confidence lower than 5
    2. Has a tag
    3. Receives a synapse
    4. Makes a synapse
    Otherwise return the last node.
    """
    if not sequence:
        raise Exception('No nodes ahead!')

    if 1 == len(sequence):
        return sequence[0]

    cursor = connection.cursor()
    cursor.execute('''
    SELECT t.id, t.confidence, tc.relation_id, tci.relation_id
    FROM treenode t
         LEFT OUTER JOIN treenode_connector tc ON (tc.treenode_id = t.id)
         LEFT OUTER JOIN treenode_class_instance tci ON (tci.treenode_id = t.id)
    WHERE t.id IN (%s)
    ''' % ",".join(str(x) for x in sequence))

    nodes = {row[0]: row for row in cursor.fetchall()}
    for nodeID in sequence:
        if nodeID in nodes:
            props = nodes[nodeID]
            # [1]: confidence
            # [2]: a treenode_connector.relation_id, e.g. presynaptic_to or postsynaptic_to
            # [3]: a treenode_class_instance.relation_id, e.g. labeled_as
            # 2 and 3 may be None
            if props[1] < 5 or props[2] or props[3]:
                return nodeID
        else:
            raise Exception('Nodes of this skeleton changed while inspecting them.')

    return sequence[-1]


@requires_user_role([UserRole.Annotate, UserRole.Browse])
def find_previous_branchnode_or_root(request, project_id=None):
    try:
        tnid = int(request.POST['tnid'])
        alt = 1 == int(request.POST['alt'])
        skid = Treenode.objects.get(pk=tnid).skeleton_id
        graph = _skeleton_as_graph(skid)
        # Travel upstream until finding a parent node with more than one child 
        # or reaching the root node
        seq = [] # Does not include the starting node tnid
        while True:
            parents = graph.predecessors(tnid)
            if parents: # list of parents is not empty
                tnid = parents[0] # Can ony have one parent
                seq.append(tnid)
                if 1 != len(graph.successors(tnid)):
                    break # Found a branch node
            else:
                break # Found the root node

        if seq and alt:
            tnid = _find_first_interesting_node(seq)

        return HttpResponse(json.dumps(_fetch_location(tnid)))
    except Exception as e:
        raise Exception('Could not obtain previous branch node or root:' + str(e))


@requires_user_role([UserRole.Annotate, UserRole.Browse])
def find_next_branchnode_or_end(request, project_id=None):
    try:
        tnid = int(request.POST['tnid'])
        shift = 1 == int(request.POST['shift'])
        alt = 1 == int(request.POST['alt'])
        skid = Treenode.objects.get(pk=tnid).skeleton_id
        graph = _skeleton_as_graph(skid)

        children = graph.successors(tnid)
        if len(children) > 1:
            # Choose one of the children:
            # The closest to 0,0,0 or the furthest if shift is down
            sqDist = 0 if shift else float('inf')
            for t in Treenode.objects.filter(parent_id=tnid):
                p = t.location
                d = pow(p.x, 2) + pow(p.y, 2) + pow(p.z, 2)
                if (shift and d > sqDist) or (not shift and d < sqDist):
                    sqDist = d
                    tnid = t.id

        # Travel downstream until finding a child node with more than one child
        # or reaching an end node
        seq = [] # Does not include the starting node tnid
        while True:
            children = graph.successors(tnid)
            if 1 == len(children):
                tnid = children[0]
                seq.append(tnid)
            else:
                break # Found an end node or a branch node

        if seq and alt:
            tnid = _find_first_interesting_node(seq)

        return HttpResponse(json.dumps(_fetch_location(tnid)))
    except Exception as e:
        raise Exception('Could not obtain next branch node or root:' + str(e))

@requires_user_role([UserRole.Browse])
def user_info(request, project_id=None):
    treenode_id = int(request.POST['treenode_id'])
    ts = Treenode.objects.filter(pk=treenode_id).select_related('user', 'editor')
    if not ts:
        ts = Connector.objects.filter(pk=treenode_id).select_related('user', 'editor')
        if not ts:
            return HttpResponse(json.dumps({'error': 'Object #%s is not a treenode or a connector' % treenode_id}))
    t = ts[0]
    reviewer = None
    review_time = None
    if -1 != t.reviewer_id:
        reviewer = User.objects.filter(pk=t.reviewer_id).values('username', 'first_name', 'last_name')[0]
        review_time = str(datetime.date(t.review_time))
    return HttpResponse(json.dumps({'user': {'username': t.user.username,
                                             'first_name': t.user.first_name,
                                             'last_name': t.user.last_name},
                                    'creation_time': str(datetime.date(t.creation_time)),
                                    'editor': {'username': t.editor.username,
                                               'first_name': t.editor.first_name,
                                               'last_name': t.editor.last_name},
                                    'edition_time': str(datetime.date(t.edition_time)),
                                    'reviewer': reviewer,
                                    'review_time': review_time}))


